from typing import List, Optional, Sequence
from collections import defaultdict

from trilogy.core.models import (
    Grain,
    QueryDatasource,
    SourceType,
    Concept,
    Environment,
    Conditional,
    UnnestJoin,
    Datasource,
    Comparison,
    Parenthetical,
    LooseConceptList,
)
from trilogy.core.enums import Purpose, JoinType, PurposeLineage, Granularity
from trilogy.utility import unique
from dataclasses import dataclass


def concept_list_to_grain(
    inputs: List[Concept], parent_sources: Sequence[QueryDatasource | Datasource]
) -> Grain:
    candidates = [
        c
        for c in inputs
        if c.purpose == Purpose.KEY and c.granularity != Granularity.SINGLE_ROW
    ]
    for x in inputs:
        if x.granularity == Granularity.SINGLE_ROW:
            continue
        if x.purpose == Purpose.PROPERTY and not any(
            [key in candidates for key in (x.keys or [])]
        ):
            candidates.append(x)
        elif x.purpose == Purpose.CONSTANT:
            candidates.append(x)
        elif x.purpose == Purpose.METRIC:
            # metrics that were previously calculated must be included in grain
            if any([x in parent.output_concepts for parent in parent_sources]):
                candidates.append(x)

    return Grain(components=candidates)


def resolve_concept_map(
    inputs: List[QueryDatasource],
    targets: List[Concept],
    inherited_inputs: List[Concept],
    full_joins: List[Concept] | None = None,
) -> dict[str, set[Datasource | QueryDatasource | UnnestJoin]]:
    targets = targets or []
    concept_map: dict[str, set[Datasource | QueryDatasource | UnnestJoin]] = (
        defaultdict(set)
    )
    full_addresses = {c.address for c in full_joins} if full_joins else set()
    for input in inputs:
        for concept in input.output_concepts:
            if concept.address not in input.non_partial_concept_addresses:
                continue
            if concept.address not in [t.address for t in inherited_inputs]:
                continue
            if concept.address in full_addresses:
                concept_map[concept.address].add(input)
            elif concept.address not in concept_map:
                concept_map[concept.address].add(input)

    # second loop, include partials
    for input in inputs:
        for concept in input.output_concepts:
            if concept.address not in [t.address for t in inherited_inputs]:
                continue
            if len(concept_map.get(concept.address, [])) == 0:
                concept_map[concept.address].add(input)
    # this adds our new derived metrics, which are not created in this CTE
    for target in targets:
        if target not in inherited_inputs:
            # an empty source means it is defined in this CTE
            concept_map[target.address] = set()
    return concept_map


def get_all_parent_partial(all_concepts: List[Concept], parents: List["StrategyNode"]):
    return [
        c
        for c in all_concepts
        if len([c.address in [x.address for x in p.partial_concepts] for p in parents])
        >= 1
        and all([c.address in [x.address for x in p.partial_concepts] for p in parents])
    ]


class StrategyNode:
    source_type = SourceType.ABSTRACT

    def __init__(
        self,
        input_concepts: List[Concept],
        output_concepts: List[Concept],
        environment: Environment,
        g,
        whole_grain: bool = False,
        parents: List["StrategyNode"] | None = None,
        partial_concepts: List[Concept] | None = None,
        depth: int = 0,
        conditions: Conditional | Comparison | Parenthetical | None = None,
        force_group: bool | None = None,
        grain: Optional[Grain] = None,
    ):
        self.input_concepts: List[Concept] = (
            unique(input_concepts, "address") if input_concepts else []
        )
        self.input_lcl = LooseConceptList(concepts=self.input_concepts)
        self.output_concepts: List[Concept] = unique(output_concepts, "address")
        self.output_lcl = LooseConceptList(concepts=self.output_concepts)

        self.environment = environment
        self.g = g
        self.whole_grain = whole_grain
        self.parents = parents or []
        self.resolution_cache: Optional[QueryDatasource] = None
        self.partial_concepts = partial_concepts or get_all_parent_partial(
            self.output_concepts, self.parents
        )
        self.partial_lcl = LooseConceptList(concepts=self.partial_concepts)
        self.depth = depth
        self.conditions = conditions
        self.grain = grain
        self.force_group = force_group
        self.tainted = False
        for parent in self.parents:
            if not parent:
                raise SyntaxError("Unresolvable parent")

    def add_output_concept(self, concept: Concept):
        self.output_concepts.append(concept)
        self.output_lcl = LooseConceptList(concepts=self.output_concepts)
        self.rebuild_cache()

    @property
    def logging_prefix(self) -> str:
        return "\t" * self.depth

    @property
    def all_concepts(self) -> list[Concept]:
        return [*self.output_concepts]

    @property
    def all_used_concepts(self) -> list[Concept]:
        return [*self.input_concepts]

    def __repr__(self):
        concepts = self.all_concepts
        contents = ",".join(sorted([c.address for c in concepts]))
        return f"{self.__class__.__name__}<{contents}>"

    def _resolve(self) -> QueryDatasource:
        parent_sources = [p.resolve() for p in self.parents]

        # if conditional:
        #     for condition in conditions[1:]:
        #         conditional += condition
        grain = Grain(components=self.output_concepts)
        source_map = resolve_concept_map(
            parent_sources, self.output_concepts, self.input_concepts
        )
        return QueryDatasource(
            input_concepts=self.input_concepts,
            output_concepts=self.output_concepts,
            datasources=parent_sources,
            source_type=self.source_type,
            source_map=source_map,
            joins=[],
            grain=grain,
            condition=self.conditions,
            partial_concepts=self.partial_concepts,
            force_group=self.force_group,
        )

    def rebuild_cache(self) -> QueryDatasource:
        self.tainted = True
        if not self.resolution_cache:
            return self.resolve()
        self.resolution_cache = None
        return self.resolve()

    def resolve(self) -> QueryDatasource:
        if self.resolution_cache:
            return self.resolution_cache
        qds = self._resolve()
        self.resolution_cache = qds
        return qds


@dataclass
class NodeJoin:
    left_node: StrategyNode
    right_node: StrategyNode
    concepts: List[Concept]
    join_type: JoinType
    filter_to_mutual: bool = False

    def __post_init__(self):
        final_concepts = []
        for concept in self.concepts:
            include = True
            for ds in [self.left_node, self.right_node]:
                if concept.address not in [c.address for c in ds.all_concepts]:
                    if self.filter_to_mutual:
                        include = False
                    else:
                        raise SyntaxError(
                            f"Invalid join, missing {concept} on {str(ds)}, have"
                            f" {[c.address for c in ds.all_concepts]}"
                        )
            if include:
                final_concepts.append(concept)
        if not final_concepts and self.concepts:
            # if one datasource only has constants
            # we can join on 1=1
            for ds in [self.left_node, self.right_node]:
                if all(
                    [c.derivation == PurposeLineage.CONSTANT for c in ds.all_concepts]
                ):
                    self.concepts = []
                    return

            left_keys = [c.address for c in self.left_node.all_concepts]
            right_keys = [c.address for c in self.right_node.all_concepts]
            match_concepts = [c.address for c in self.concepts]
            raise SyntaxError(
                "No mutual join keys found between"
                f" {self.left_node} and"
                f" {self.right_node}, left_keys {left_keys},"
                f" right_keys {right_keys},"
                f" provided join concepts {match_concepts}"
            )
        self.concepts = final_concepts

    @property
    def unique_id(self) -> str:
        nodes = sorted([self.left_node, self.right_node], key=lambda x: str(x))
        return str(nodes) + self.join_type.value

    def __str__(self):
        return (
            f"{self.join_type.value} JOIN {self.left_node} and"
            f" {self.right_node} on"
            f" {','.join([str(k) for k in self.concepts])}"
        )



from .filter_node import gen_filter_node
from .window_node import gen_window_node
from .group_node import gen_group_node
from .group_to_node import gen_group_to_node
from .basic_node import gen_basic_node
from .select_node import gen_select_node
from .unnest_node import gen_unnest_node
from .merge_node import gen_merge_node
from .rowset_node import gen_rowset_node
from .multiselect_node import gen_multiselect_node
from .concept_merge import gen_concept_merge_node

__all__ = [
    "gen_filter_node",
    "gen_window_node",
    "gen_group_node",
    "gen_select_node",
    "gen_basic_node",
    "gen_unnest_node",
    "gen_merge_node",
    "gen_group_to_node",
    "gen_rowset_node",
    "gen_multiselect_node",
    "gen_concept_merge_node",
]



from itertools import combinations
from typing import List, Optional

from trilogy.core.enums import PurposeLineage
from trilogy.core.models import (
    Concept,
    Environment,
    Grain,
    LooseConceptList,
    Datasource,
)
from trilogy.core.processing.nodes import (
    StrategyNode,
    SelectNode,
    MergeNode,
    GroupNode,
    ConstantNode,
)
from trilogy.core.exceptions import NoDatasourceException
import networkx as nx
from trilogy.core.graph_models import concept_to_node, datasource_to_node
from trilogy.constants import logger
from trilogy.core.processing.utility import padding
from dataclasses import dataclass

LOGGER_PREFIX = "[GEN_SELECT_NODE]"


@dataclass
class DatasourceMatch:
    key: str
    datasource: Datasource
    matched: LooseConceptList
    partial: LooseConceptList

    def __repr__(self):
        return f"DatasourceMatch({self.key}, {self.datasource.identifier}, {str(self.matched)}, {str(self.partial)})"


def dm_to_strategy_node(
    dm: DatasourceMatch,
    target_grain: Grain,
    environment: Environment,
    g: nx.DiGraph,
    depth: int,
    accept_partial: bool = False,
) -> StrategyNode:
    datasource = dm.datasource
    if target_grain and target_grain.issubset(datasource.grain):
        if all([x in dm.matched for x in target_grain.components]):
            force_group = False
        # if we are not returning the grain
        # we have to group
        else:
            logger.info(
                f"{padding(depth)}{LOGGER_PREFIX} not all grain components are in output {str(dm.matched)}, group to actual grain"
            )
            force_group = True
    elif all([x in dm.matched for x in datasource.grain.components]):
        logger.info(
            f"{padding(depth)}{LOGGER_PREFIX} query output includes all grain components, no reason to group further"
        )
        force_group = False
    else:
        logger.info(
            f"{padding(depth)}{LOGGER_PREFIX} target grain is not subset of datasource grain {datasource.grain}, required to group"
        )
        force_group = True
    bcandidate: StrategyNode = SelectNode(
        input_concepts=[c.concept for c in datasource.columns],
        output_concepts=dm.matched.concepts,
        environment=environment,
        g=g,
        parents=[],
        depth=depth,
        partial_concepts=dm.partial.concepts,
        accept_partial=accept_partial,
        datasource=datasource,
        grain=Grain(components=dm.matched.concepts),
    )
    # we need to nest the group node one further
    if force_group is True:
        candidate: StrategyNode = GroupNode(
            output_concepts=dm.matched.concepts,
            input_concepts=dm.matched.concepts,
            environment=environment,
            g=g,
            parents=[bcandidate],
            depth=depth,
            partial_concepts=bcandidate.partial_concepts,
        )
    else:
        candidate = bcandidate
    return candidate


def gen_select_nodes_from_tables_v2(
    mandatory_concept: Concept,
    all_concepts: List[Concept],
    g: nx.DiGraph,
    environment: Environment,
    depth: int,
    target_grain: Grain,
    accept_partial: bool = False,
) -> tuple[bool, list[Concept], list[StrategyNode]]:
    # if we have only constants
    # we don't need a table
    # so verify nothing, select node will render
    all_lcl = LooseConceptList(concepts=all_concepts)
    if all([c.derivation == PurposeLineage.CONSTANT for c in all_lcl]):
        logger.info(
            f"{padding(depth)}{LOGGER_PREFIX} All concepts {[x.address for x in all_lcl]} are constants, returning constant node"
        )
        return (
            True,
            all_lcl.concepts,
            [
                ConstantNode(
                    output_concepts=all_lcl.concepts,
                    input_concepts=[],
                    environment=environment,
                    g=g,
                    parents=[],
                    depth=depth,
                    # no partial for constants
                    partial_concepts=[],
                    force_group=False,
                )
            ],
        )
    # otherwise, we need to look for a table
    nodes_to_find = [concept_to_node(x.with_default_grain()) for x in all_lcl.concepts]
    matches: dict[str, DatasourceMatch] = {}
    for k, datasource in environment.datasources.items():
        matched = []
        matched_paths = []
        for idx, req_concept in enumerate(nodes_to_find):
            try:
                path = nx.shortest_path(
                    g,
                    source=datasource_to_node(datasource),
                    target=req_concept,
                )
                ds_valid = (
                    sum(
                        [
                            1 if g.nodes[node]["type"] == "datasource" else 0
                            for node in path
                        ]
                    )
                    == 1
                )
                address_valid = (
                    sum(
                        [
                            (
                                1
                                if g.nodes[node]["type"] == "concept"
                                and g.nodes[node]["concept"].address
                                != all_lcl.concepts[idx].address
                                else 0
                            )
                            for node in path
                        ]
                    )
                    == 0
                )
                if ds_valid and address_valid:
                    matched_paths.append(path)
                    matched.append(all_lcl.concepts[idx])
            except nx.NodeNotFound:
                continue
            except nx.exception.NetworkXNoPath:
                continue
        dm = DatasourceMatch(
            key=k,
            datasource=datasource,
            matched=LooseConceptList(concepts=matched),
            partial=LooseConceptList(
                concepts=[
                    c.concept
                    for c in datasource.columns
                    if not c.is_complete and c.concept.address in all_lcl
                ]
            ),
        )
        if not matched:
            continue
        if mandatory_concept.address not in dm.matched:
            continue
        if not accept_partial and dm.partial.addresses:
            continue
        matches[k] = dm
    found: set[str] = set()
    all_found = False
    all_checked = False
    to_return: list[StrategyNode] = []
    if not matches:
        return False, [], []
    while not all_found and not all_checked:
        final_key: str = max(
            matches,
            key=lambda x: len(
                [m for m in matches[x].matched.addresses if m not in found]
            )
            - 0.1 * len(matches[x].partial.addresses),
        )
        final: DatasourceMatch = matches[final_key]
        candidate = dm_to_strategy_node(
            final,
            target_grain=Grain(
                components=[
                    x for x in target_grain.components if x.address in final.matched
                ]
            ),
            environment=environment,
            g=g,
            depth=depth,
            accept_partial=accept_partial,
        )
        to_return.append(candidate)
        del matches[final_key]
        found = found.union(final.matched.addresses)
        all_found = all_lcl.addresses.issubset(found)
        all_checked = len(matches) == 0
    return all_found, [x for x in all_concepts if x.address in found], to_return


def gen_select_node_from_table(
    target_concept: Concept,
    all_concepts: List[Concept],
    g: nx.DiGraph,
    environment: Environment,
    depth: int,
    target_grain: Grain,
    accept_partial: bool = False,
) -> Optional[StrategyNode]:
    # if we have only constants
    # we don't need a table
    # so verify nothing, select node will render
    all_lcl = LooseConceptList(concepts=all_concepts)
    if all([c.derivation == PurposeLineage.CONSTANT for c in all_concepts]):
        logger.info(
            f"{padding(depth)}{LOGGER_PREFIX} All concepts {[x.address for x in all_concepts]} are constants, returning constant node"
        )
        return ConstantNode(
            output_concepts=all_concepts,
            input_concepts=[],
            environment=environment,
            g=g,
            parents=[],
            depth=depth,
            # no partial for constants
            partial_concepts=[],
            force_group=False,
        )
    candidates: dict[str, StrategyNode] = {}
    scores: dict[str, int] = {}
    # otherwise, we need to look for a table
    nodes_to_find = [concept_to_node(x.with_default_grain()) for x in all_concepts]
    for datasource in environment.datasources.values():
        all_found = True
        for idx, req_concept in enumerate(nodes_to_find):
            try:
                path = nx.shortest_path(
                    g,
                    source=datasource_to_node(datasource),
                    target=req_concept,
                )
            except nx.NodeNotFound as e:
                # just to provide better error
                ncandidates = [
                    datasource_to_node(datasource),
                    req_concept,
                ]
                for ncandidate in ncandidates:
                    try:
                        g.nodes[ncandidate]
                    except KeyError:
                        raise SyntaxError(
                            "Could not find node for {}".format(ncandidate)
                        )
                raise e
            except nx.exception.NetworkXNoPath:
                all_found = False
                break
            # 2023-10-18 - more strict condition then below
            # 2023-10-20 - we need this to get through abstract concepts
            # but we may want to add a filter to avoid using this for anything with lineage
            # if len(path) != 2:
            #     all_found = False
            #     break
            if len([p for p in path if g.nodes[p]["type"] == "datasource"]) != 1:
                all_found = False
                break
            for node in path:
                if g.nodes[node]["type"] == "datasource":
                    continue
                if g.nodes[node]["concept"].address == all_concepts[idx].address:
                    continue
                all_found = False
                break

        if not all_found:
            # skip to next node
            continue
        partial_concepts = [
            c.concept
            for c in datasource.columns
            if not c.is_complete and c.concept in all_lcl
        ]
        partial_lcl = LooseConceptList(concepts=partial_concepts)
        if not accept_partial and target_concept in partial_lcl:
            continue
        logger.info(
            f"{padding(depth)}{LOGGER_PREFIX} target grain is {str(target_grain)}"
        )
        if target_grain and target_grain.issubset(datasource.grain):

            if all([x in all_lcl for x in target_grain.components]):
                force_group = False
            # if we are not returning the grain
            # we have to group
            else:
                logger.info(
                    f"{padding(depth)}{LOGGER_PREFIX} not all grain components are in output {str(all_lcl)}, group to actual grain"
                )
                force_group = True
        elif all([x in all_lcl for x in datasource.grain.components]):
            logger.info(
                f"{padding(depth)}{LOGGER_PREFIX} query output includes all grain components, no reason to group further"
            )
            force_group = False
        else:
            logger.info(
                f"{padding(depth)}{LOGGER_PREFIX} target grain is not subset of datasource grain {datasource.grain}, required to group"
            )
            force_group = True

        bcandidate: StrategyNode = SelectNode(
            input_concepts=[c.concept for c in datasource.columns],
            output_concepts=all_concepts,
            environment=environment,
            g=g,
            parents=[],
            depth=depth,
            partial_concepts=[c for c in all_concepts if c in partial_lcl],
            accept_partial=accept_partial,
            datasource=datasource,
            grain=Grain(components=all_concepts),
        )
        # we need to nest the group node one further
        if force_group is True:
            candidate: StrategyNode = GroupNode(
                output_concepts=all_concepts,
                input_concepts=all_concepts,
                environment=environment,
                g=g,
                parents=[bcandidate],
                depth=depth,
                partial_concepts=bcandidate.partial_concepts,
            )
        else:
            candidate = bcandidate
        logger.info(
            f"{padding(depth)}{LOGGER_PREFIX} found select node with {datasource.identifier}, returning {candidate.output_lcl}"
        )
        candidates[datasource.identifier] = candidate
        scores[datasource.identifier] = -len(partial_concepts)
    if not candidates:
        return None
    final = max(candidates, key=lambda x: scores[x])
    return candidates[final]


def gen_select_nodes_from_tables(
    local_optional: List[Concept],
    depth: int,
    concept: Concept,
    environment: Environment,
    g: nx.DiGraph,
    accept_partial: bool,
    all_concepts: List[Concept],
) -> tuple[bool, list[Concept], list[StrategyNode]]:
    parents: List[StrategyNode] = []
    found: List[Concept] = []
    logger.info(
        f"{padding(depth)}{LOGGER_PREFIX} looking for multiple sources that can satisfy"
    )
    all_found = False
    unreachable: list[str] = []
    # first pass
    for opt_con in local_optional:
        ds = gen_select_node_from_table(
            concept,
            [concept, opt_con],
            g=g,
            environment=environment,
            depth=depth + 1,
            accept_partial=accept_partial,
            target_grain=Grain(components=all_concepts),
        )
        if not ds:
            unreachable.append(opt_con.address)
    all_found = False
    for x in reversed(range(1, len(local_optional) + 1)):
        if all_found:
            break
        for combo in combinations(local_optional, x):
            if all_found:
                break
            # filter to just the original ones we need to get
            local_combo = [
                x for x in combo if x not in found and x.address not in unreachable
            ]
            # skip if nothing new in this combo
            if not local_combo:
                continue
            # include core concept as join
            all_concepts = [concept, *local_combo]

            ds = gen_select_node_from_table(
                concept,
                all_concepts,
                g=g,
                environment=environment,
                depth=depth + 1,
                accept_partial=accept_partial,
                target_grain=Grain(components=all_concepts),
            )
            if ds:
                logger.info(
                    f"{padding(depth)}{LOGGER_PREFIX} found a source with {[x.address for x in all_concepts]}"
                )
                parents.append(ds)
                found += [x for x in ds.output_concepts if x != concept]
                if {x.address for x in found} == {c.address for c in local_optional}:
                    logger.info(
                        f"{padding(depth)}{LOGGER_PREFIX} found all optional {[c.address for c in local_optional]}"
                    )
                    all_found = True
    return all_found, found, parents


def gen_select_node(
    concept: Concept,
    local_optional: List[Concept],
    environment: Environment,
    g,
    depth: int,
    accept_partial: bool = False,
    fail_if_not_found: bool = True,
    accept_partial_optional: bool = True,
    target_grain: Grain | None = None,
) -> StrategyNode | None:
    all_concepts = [concept] + local_optional
    all_lcl = LooseConceptList(concepts=all_concepts)
    materialized_lcl = LooseConceptList(
        concepts=[
            x
            for x in all_concepts
            if x.address in [z.address for z in environment.materialized_concepts]
            or x.derivation == PurposeLineage.CONSTANT
        ]
    )
    if not target_grain:
        target_grain = Grain()
        for ac in all_concepts:
            target_grain += ac.grain
    if materialized_lcl != all_lcl:
        logger.info(
            f"{padding(depth)}{LOGGER_PREFIX} Skipping select node generation for {concept.address} "
            f" as it + optional (looking for all {all_lcl}) includes non-materialized concepts {all_lcl.difference(materialized_lcl)} vs materialized: {materialized_lcl}"
        )
        if fail_if_not_found:
            raise NoDatasourceException(f"No datasource exists for {concept}")
        return None

    ds: StrategyNode | None = None

    # attempt to select all concepts from table
    ds = gen_select_node_from_table(
        concept,
        [concept] + local_optional,
        g=g,
        environment=environment,
        depth=depth,
        accept_partial=accept_partial,
        target_grain=target_grain,
    )
    if ds:
        logger.info(
            f"{padding(depth)}{LOGGER_PREFIX} Found select node with all target concepts, force group is {ds.force_group}, target grain {target_grain}"
        )
        return ds
    # if we cannot find a match
    all_found, found, parents = gen_select_nodes_from_tables_v2(
        concept, all_concepts, g, environment, depth, target_grain, accept_partial
    )
    if parents and (all_found or accept_partial_optional):
        if all_found:
            logger.info(
                f"{padding(depth)}{LOGGER_PREFIX} found all optional {[c.address for c in local_optional]} via joins"
            )
        else:
            logger.info(
                f"{padding(depth)}{LOGGER_PREFIX} found some optional, returning"
            )
        all_partial = [
            c
            for c in all_concepts
            if all(
                [c.address in [x.address for x in p.partial_concepts] for p in parents]
            )
        ]
        force_group = False
        for candidate in parents:
            if candidate.grain and not candidate.grain.issubset(target_grain):
                force_group = True
        if len(parents) == 1:
            candidate = parents[0]
        else:
            candidate = MergeNode(
                output_concepts=[concept] + found,
                input_concepts=[concept] + found,
                environment=environment,
                g=g,
                parents=parents,
                depth=depth,
                partial_concepts=all_partial,
                grain=sum([x.grain for x in parents if x.grain], Grain()),
            )
        candidate.depth += 1
        source_grain = candidate.grain
        if force_group:
            logger.info(
                f"{padding(depth)}{LOGGER_PREFIX} datasource grain {source_grain} does not match target grain {target_grain} for select, adding group node"
            )
            return GroupNode(
                output_concepts=candidate.output_concepts,
                input_concepts=candidate.output_concepts,
                environment=environment,
                g=g,
                parents=[candidate],
                depth=depth,
                partial_concepts=candidate.partial_concepts,
            )
        return candidate

    if not accept_partial_optional:
        return None
    ds = gen_select_node_from_table(
        concept,
        [concept],
        g=g,
        environment=environment,
        depth=depth,
        accept_partial=accept_partial,
        target_grain=Grain(components=[concept]),
    )

    if not ds and fail_if_not_found:
        raise NoDatasourceException(f"No datasource exists for {concept}")
    return ds



from typing import List, Optional

from trilogy.core.models import Concept, Environment, Datasource, Conditional
from trilogy.core.processing.nodes import MergeNode, History
import networkx as nx
from trilogy.core.graph_models import concept_to_node, datasource_to_node
from trilogy.core.processing.utility import PathInfo
from trilogy.constants import logger
from trilogy.utility import unique
from trilogy.core.exceptions import AmbiguousRelationshipResolutionException
from trilogy.core.processing.utility import padding
from trilogy.core.processing.graph_utils import extract_mandatory_subgraphs

LOGGER_PREFIX = "[GEN_MERGE_NODE]"


def reduce_path_concepts(paths, g) -> set[str]:
    concept_nodes: List[Concept] = []
    # along our path, find all the concepts required
    for _, value in paths.items():
        concept_nodes += [g.nodes[v]["concept"] for v in value if v.startswith("c~")]
    final: List[Concept] = unique(concept_nodes, "address")
    return set([x.address for x in final])


def identify_ds_join_paths(
    all_concepts: List[Concept],
    g,
    datasource: Datasource,
    accept_partial: bool = False,
    fail: bool = False,
) -> PathInfo | None:
    all_found = True
    any_direct_found = False
    paths = {}
    for bitem in all_concepts:
        item = bitem.with_default_grain()
        target_node = concept_to_node(item)
        try:
            path = nx.shortest_path(
                g,
                source=datasource_to_node(datasource),
                target=target_node,
            )
            paths[target_node] = path
            if sum([1 for x in path if x.startswith("ds~")]) == 1:
                any_direct_found = True
        except nx.exception.NodeNotFound:
            # TODO: support Verbose logging mode configuration and reenable these
            all_found = False
            if fail:
                raise
            return None
        except nx.exception.NetworkXNoPath:
            all_found = False
            if fail:
                raise
            return None
    if all_found and any_direct_found:
        partial = [
            c.concept
            for c in datasource.columns
            if not c.is_complete
            and c.concept.address in [x.address for x in all_concepts]
        ]
        if partial and not accept_partial:
            return None
        # join_candidates.append({"paths": paths, "datasource": datasource})
        return PathInfo(
            paths=paths,
            datasource=datasource,
            reduced_concepts=reduce_path_concepts(paths, g),
            concept_subgraphs=extract_mandatory_subgraphs(paths, g),
        )  # {"paths": paths, "datasource": datasource}
    return None


def gen_merge_node(
    all_concepts: List[Concept],
    g: nx.DiGraph,
    environment: Environment,
    depth: int,
    source_concepts,
    accept_partial: bool = False,
    history: History | None = None,
    conditions: Conditional | None = None,
) -> Optional[MergeNode]:
    join_candidates: List[PathInfo] = []
    # anchor on datasources
    final_all_concepts = []
    # implicit_upstream = {}
    for x in all_concepts:
        # if x.derivation in (PurposeLineage.AGGREGATE, PurposeLineage.BASIC):
        #     final_all_concepts +=resolve_function_parent_concepts(x)
        # elif x.derivation == PurposeLineage.FILTER:
        #     final_all_concepts +=resolve_filter_parent_concepts(x)
        # else:
        #     final_all_concepts.append(x)
        final_all_concepts.append(x)
    for datasource in environment.datasources.values():
        path = identify_ds_join_paths(final_all_concepts, g, datasource, accept_partial)
        if path and path.reduced_concepts:
            join_candidates.append(path)
    join_candidates.sort(key=lambda x: sum([len(v) for v in x.paths.values()]))
    if not join_candidates:
        return None
    for join_candidate in join_candidates:
        logger.info(
            f"{padding(depth)}{LOGGER_PREFIX} Join candidate: {join_candidate.paths}"
        )
    join_additions: List[set[str]] = []
    for candidate in join_candidates:
        join_additions.append(candidate.reduced_concepts)
    if not all(
        [x.issubset(y) or y.issubset(x) for x in join_additions for y in join_additions]
    ):
        raise AmbiguousRelationshipResolutionException(
            f"Ambiguous concept join resolution - possible paths =  {join_additions}. Include an additional concept to disambiguate",
            join_additions,
        )
    if not join_candidates:
        logger.info(
            f"{padding(depth)}{LOGGER_PREFIX} No additional join candidates could be found"
        )
        return None
    shortest: PathInfo = sorted(join_candidates, key=lambda x: len(x.reduced_concepts))[
        0
    ]
    logger.info(f"{padding(depth)}{LOGGER_PREFIX} final path is {shortest.paths}")
    # logger.info(f'{padding(depth)}{LOGGER_PREFIX} final reduced concepts are {shortest.concs}')
    parents = []
    for graph in shortest.concept_subgraphs:
        logger.info(
            f"{padding(depth)}{LOGGER_PREFIX} fetching subgraph {[c.address for c in graph]}"
        )
        parent = source_concepts(
            mandatory_list=graph,
            environment=environment,
            g=g,
            depth=depth + 1,
            history=history,
        )
        if not parent:
            logger.info(
                f"{padding(depth)}{LOGGER_PREFIX} Unable to instantiate target subgraph"
            )
            return None
        parents.append(parent)

    return MergeNode(
        input_concepts=[environment.concepts[x] for x in shortest.reduced_concepts],
        output_concepts=all_concepts,
        environment=environment,
        g=g,
        parents=parents,
        depth=depth,
        conditions=conditions,
    )



from trilogy.core.models import Concept, DataType, Function
from trilogy.core.enums import Purpose, FunctionType
from trilogy.core.constants import ALL_ROWS_CONCEPT, INTERNAL_NAMESPACE


DEFAULT_CONCEPTS = {
    ALL_ROWS_CONCEPT: Concept(
        name=ALL_ROWS_CONCEPT,
        namespace=INTERNAL_NAMESPACE,
        datatype=DataType.INTEGER,
        purpose=Purpose.CONSTANT,
        lineage=Function(
            operator=FunctionType.CONSTANT,
            arguments=[1],
            output_datatype=DataType.INTEGER,
            output_purpose=Purpose.CONSTANT,
        ),
    ),
    "concept_name": Concept(
        name="concept_name",
        namespace=INTERNAL_NAMESPACE,
        datatype=DataType.STRING,
        purpose=Purpose.KEY,
    ),
    "datasource": Concept(
        name="datasource",
        namespace=INTERNAL_NAMESPACE,
        datatype=DataType.STRING,
        purpose=Purpose.KEY,
    ),
    "query_text": Concept(
        name="query_text",
        namespace=INTERNAL_NAMESPACE,
        datatype=DataType.STRING,
        purpose=Purpose.KEY,
    ),
}



from trilogy.core.graph_models import (
    ReferenceGraph,
    concept_to_node,
    datasource_to_node,
)
from trilogy.core.models import Environment
from trilogy.core.enums import PurposeLineage


def generate_graph(
    environment: Environment,
) -> ReferenceGraph:
    g = ReferenceGraph()

    # add all parsed concepts
    for _, concept in environment.concepts.items():
        g.add_node(concept)
        # if we have sources, recursively add them
        if concept.sources:
            node_name = concept_to_node(concept)
            for source in concept.sources:
                generic = source.with_default_grain()
                g.add_edge(generic, node_name)
                if concept.derivation == PurposeLineage.MERGE:
                    g.add_edge(node_name, generic)
    for _, dataset in environment.datasources.items():
        node = datasource_to_node(dataset)
        g.add_node(dataset, type="datasource", datasource=dataset)
        for concept in dataset.concepts:
            g.add_edge(node, concept)
            g.add_edge(concept, node)
            # if there is a key on a table at a different grain
            # add an FK edge to the canonical source, if it exists
            # for example, order ID on order product table
            g.add_edge(concept, concept.with_default_grain())
            g.add_edge(concept.with_default_grain(), concept)
    return g



class DialectConfig:
    pass


class DuckDBConfig(DialectConfig):
    def __init__(self, path: str | None = None):
        self.path = path

    def connection_string(self) -> str:
        if not self.path:
            return "duckdb:///:memory:"
        return f"duckdb:///{self.path}"


class PostgresConfig(DialectConfig):
    def __init__(
        self, host: str, port: int, username: str, password: str, database: str
    ):
        self.host = host
        self.port = port
        self.username = username
        self.password = password
        self.database = database

    def connection_string(self) -> str:
        return f"postgresql://{self.username}:{self.password}@{self.host}:{self.port}"


class SQLServerConfig(DialectConfig):
    def __init__(
        self, host: str, port: int, username: str, password: str, database: str
    ):
        self.host = host
        self.port = port
        self.username = username
        self.password = password
        self.database = database

    def connection_string(self) -> str:
        return f"sqlserver//{self.username}:{self.password}@{self.host}:{self.port}"


class SnowflakeConfig(DialectConfig):
    def __init__(
        self,
        account: str,
        username: str,
        password: str,
    ):
        self.account = account
        self.username = username
        self.password = password

    def connection_string(self) -> str:
        return f"snowflake://{self.username}:{self.password}@{self.account}"



import hashlib
from typing import List, Any, Union, Callable

from trilogy.constants import DEFAULT_NAMESPACE

INT_HASH_SIZE = 16


def string_to_hash(input: str) -> int:
    return (
        int(hashlib.sha256(input.encode("utf-8")).hexdigest(), 16) % 10**INT_HASH_SIZE
    )


def unique(inputs: List, property: Union[str, Callable]) -> List[Any]:
    final = []
    dedupe = set()
    if isinstance(property, str):

        def getter(x):
            return getattr(x, property, DEFAULT_NAMESPACE)

    else:
        getter = property
    for input in inputs:
        key = getter(input)
        if key in dedupe:
            continue
        dedupe.add(key)
        final.append(input)
    return final



from typing import List


from trilogy.core.models import Concept, Function
from trilogy.core.processing.nodes import UnnestNode, History


def gen_unnest_node(
    concept: Concept,
    local_optional: List[Concept],
    environment,
    g,
    depth: int,
    source_concepts,
    history: History | None = None,
) -> UnnestNode | None:
    arguments = []
    if isinstance(concept.lineage, Function):
        arguments = concept.lineage.concept_arguments
    if arguments or local_optional:
        parent = source_concepts(
            mandatory_list=arguments + local_optional,
            environment=environment,
            g=g,
            depth=depth + 1,
            history=history,
        )
        if not parent:
            return None
    return UnnestNode(
        unnest_concept=concept,
        input_concepts=arguments + local_optional,
        output_concepts=[concept] + local_optional,
        environment=environment,
        g=g,
        parents=([parent] if (arguments or local_optional) else []),
    )



from trilogy.core.models import (
    Concept,
    Environment,
    MultiSelectStatement,
)
from trilogy.core.processing.nodes import MergeNode, NodeJoin, History
from trilogy.core.processing.nodes.base_node import concept_list_to_grain, StrategyNode
from typing import List

from trilogy.core.enums import JoinType
from trilogy.constants import logger
from trilogy.core.processing.utility import padding
from trilogy.core.processing.node_generators.common import concept_to_relevant_joins
from collections import defaultdict
from itertools import combinations
from trilogy.core.enums import Purpose
from trilogy.core.processing.node_generators.common import resolve_join_order

LOGGER_PREFIX = "[GEN_MULTISELECT_NODE]"


def extra_align_joins(
    base: MultiSelectStatement, parents: List[StrategyNode]
) -> List[NodeJoin]:
    node_merge_concept_map = defaultdict(list)
    output = []
    for align in base.align.items:
        jc = align.gen_concept(base)
        if jc.purpose == Purpose.CONSTANT:
            continue
        for node in parents:
            for item in align.concepts:
                if item in node.output_lcl:
                    node_merge_concept_map[node].append(jc)

    for left, right in combinations(node_merge_concept_map.keys(), 2):
        matched_concepts = [
            x
            for x in node_merge_concept_map[left]
            if x in node_merge_concept_map[right]
        ]
        output.append(
            NodeJoin(
                left_node=left,
                right_node=right,
                concepts=matched_concepts,
                join_type=JoinType.FULL,
            )
        )
    return resolve_join_order(output)


def gen_multiselect_node(
    concept: Concept,
    local_optional: List[Concept],
    environment: Environment,
    g,
    depth: int,
    source_concepts,
    history: History | None = None,
) -> MergeNode | None:
    if not isinstance(concept.lineage, MultiSelectStatement):
        logger.info(
            f"{padding(depth)}{LOGGER_PREFIX} Cannot generate multiselect node for {concept}"
        )
        return None
    lineage: MultiSelectStatement = concept.lineage

    base_parents: List[StrategyNode] = []
    for select in lineage.selects:
        snode: StrategyNode = source_concepts(
            mandatory_list=select.output_components,
            environment=environment,
            g=g,
            depth=depth + 1,
            history=history,
        )
        if not snode:
            logger.info(
                f"{padding(depth)}{LOGGER_PREFIX} Cannot generate multiselect node for {concept}"
            )
            return None
        if select.where_clause:
            snode.conditions = select.where_clause.conditional
        merge_concepts = []
        for x in [*snode.output_concepts]:
            merge = lineage.get_merge_concept(x)
            if merge:
                snode.output_concepts.append(merge)
                merge_concepts.append(merge)
        # clear cache so QPS
        snode.rebuild_cache()
        for mc in merge_concepts:
            assert mc in snode.resolve().output_concepts
        base_parents.append(snode)

    node_joins = extra_align_joins(lineage, base_parents)
    node = MergeNode(
        input_concepts=[x for y in base_parents for x in y.output_concepts],
        output_concepts=[x for y in base_parents for x in y.output_concepts],
        environment=environment,
        g=g,
        depth=depth,
        parents=base_parents,
        node_joins=node_joins,
    )

    enrichment = set([x.address for x in local_optional])

    rowset_relevant = [
        x
        for x in lineage.derived_concepts
        if x.address == concept.address or x.address in enrichment
    ]
    additional_relevant = [
        x for x in select.output_components if x.address in enrichment
    ]
    # add in other other concepts
    for item in rowset_relevant:
        node.output_concepts.append(item)
    for item in additional_relevant:
        node.output_concepts.append(item)
    if select.where_clause:
        for item in additional_relevant:
            node.partial_concepts.append(item)

    # we need a better API for refreshing a nodes QDS
    node.resolution_cache = node._resolve()

    # assume grain to be output of select
    # but don't include anything aggregate at this point
    node.resolution_cache.grain = concept_list_to_grain(
        node.output_concepts, parent_sources=node.resolution_cache.datasources
    )
    possible_joins = concept_to_relevant_joins(additional_relevant)
    if not local_optional:
        logger.info(
            f"{padding(depth)}{LOGGER_PREFIX} no enriched required for rowset node; exiting early"
        )
        return node
    if not possible_joins:
        logger.info(
            f"{padding(depth)}{LOGGER_PREFIX} no possible joins for rowset node; exiting early"
        )
        return node
    if all(
        [x.address in [y.address for y in node.output_concepts] for x in local_optional]
    ):
        logger.info(
            f"{padding(depth)}{LOGGER_PREFIX} all enriched concepts returned from base rowset node; exiting early"
        )
        return node
    enrich_node: MergeNode = source_concepts(  # this fetches the parent + join keys
        # to then connect to the rest of the query
        mandatory_list=additional_relevant + local_optional,
        environment=environment,
        g=g,
        depth=depth + 1,
        history=history,
    )
    if not enrich_node:
        logger.info(
            f"{padding(depth)}{LOGGER_PREFIX} Cannot generate rowset enrichment node for {concept} with optional {local_optional}, returning just rowset node"
        )
        return node

    return MergeNode(
        input_concepts=enrich_node.output_concepts + node.output_concepts,
        output_concepts=node.output_concepts + local_optional,
        environment=environment,
        g=g,
        depth=depth,
        parents=[
            # this node gets the multiselect
            node,
            # this node gets enrichment
            enrich_node,
        ],
        node_joins=[
            NodeJoin(
                left_node=enrich_node,
                right_node=node,
                concepts=possible_joins,
                filter_to_mutual=False,
                join_type=JoinType.LEFT_OUTER,
            )
        ],
        partial_concepts=node.partial_concepts,
    )



from typing import List, Tuple, Dict, Set
import networkx as nx
from trilogy.core.models import (
    Datasource,
    JoinType,
    BaseJoin,
    Concept,
    QueryDatasource,
    LooseConceptList,
)

from trilogy.core.enums import Purpose, Granularity
from trilogy.core.constants import CONSTANT_DATASET
from enum import Enum
from trilogy.utility import unique
from collections import defaultdict
from logging import Logger
from pydantic import BaseModel


class NodeType(Enum):
    CONCEPT = 1
    NODE = 2


class PathInfo(BaseModel):
    paths: Dict[str, List[str]]
    datasource: Datasource
    reduced_concepts: Set[str]
    concept_subgraphs: List[List[Concept]]


def concept_to_relevant_joins(concepts: list[Concept]) -> List[Concept]:
    addresses = LooseConceptList(concepts=concepts)
    sub_props = LooseConceptList(
        concepts=[
            x for x in concepts if x.keys and all([key in addresses for key in x.keys])
        ]
    )
    final = [c for c in concepts if c not in sub_props]
    return unique(final, "address")


def padding(x: int) -> str:
    return "\t" * x


def create_log_lambda(prefix: str, depth: int, logger: Logger):
    pad = padding(depth)

    def log_lambda(msg: str):
        logger.info(f"{pad} {prefix} {msg}")

    return log_lambda


def calculate_graph_relevance(
    g: nx.DiGraph, subset_nodes: set[str], concepts: set[Concept]
) -> int:
    """Calculate the relevance of each node in a graph
    Relevance is used to prune irrelevant nodes from the graph
    """
    relevance = 0
    for node in g.nodes:
        if node not in subset_nodes:
            continue
        if not g.nodes[node]["type"] == NodeType.CONCEPT:
            continue
        concept = [x for x in concepts if x.address == node].pop()

        # a single row concept can always be crossjoined
        # therefore a graph with only single row concepts is always relevant
        if concept.granularity == Granularity.SINGLE_ROW:
            continue
        # if it's an aggregate up to an arbitrary grain, it can be joined in later
        # and can be ignored in subgraph
        if concept.purpose == Purpose.METRIC:
            if not concept.grain:
                continue
            if len(concept.grain.components) == 0:
                continue
        if concept.grain and len(concept.grain.components) > 0:
            relevance += 1
            continue
        # Added 2023-10-18 since we seemed to be strangely dropping things
        relevance += 1

    return relevance


def resolve_join_order(joins: List[BaseJoin]) -> List[BaseJoin]:
    available_aliases: set[str] = set()
    final_joins_pre = [*joins]
    final_joins = []
    while final_joins_pre:
        new_final_joins_pre: List[BaseJoin] = []
        for join in final_joins_pre:
            if not available_aliases:
                final_joins.append(join)
                available_aliases.add(join.left_datasource.identifier)
                available_aliases.add(join.right_datasource.identifier)
            elif join.left_datasource.identifier in available_aliases:
                # we don't need to join twice
                # so whatever join we found first, works
                if join.right_datasource.identifier in available_aliases:
                    continue
                final_joins.append(join)
                available_aliases.add(join.left_datasource.identifier)
                available_aliases.add(join.right_datasource.identifier)
            else:
                new_final_joins_pre.append(join)
        if len(new_final_joins_pre) == len(final_joins_pre):
            remaining = [
                join.left_datasource.identifier for join in new_final_joins_pre
            ]
            remaining_right = [
                join.right_datasource.identifier for join in new_final_joins_pre
            ]
            raise SyntaxError(
                f"did not find any new joins, available {available_aliases} remaining is {remaining + remaining_right} "
            )
        final_joins_pre = new_final_joins_pre
    return final_joins


def get_node_joins(
    datasources: List[QueryDatasource],
    grain: List[Concept],
    # concepts:List[Concept],
) -> List[BaseJoin]:
    graph = nx.Graph()
    concepts: List[Concept] = []
    for datasource in datasources:
        graph.add_node(datasource.identifier, type=NodeType.NODE)
        for concept in datasource.output_concepts:
            # we don't need to join on a concept if all of the keys exist in the grain
            # if concept.keys and all([x in grain for x in concept.keys]):
            #     continue
            concepts.append(concept)
            graph.add_node(concept.address, type=NodeType.CONCEPT)
            graph.add_edge(datasource.identifier, concept.address)

    # add edges for every constant to every datasource
    for datasource in datasources:
        for concept in datasource.output_concepts:
            if concept.granularity == Granularity.SINGLE_ROW:
                for node in graph.nodes:
                    if graph.nodes[node]["type"] == NodeType.NODE:
                        graph.add_edge(node, concept.address)

    joins: defaultdict[str, set] = defaultdict(set)
    identifier_map = {x.identifier: x for x in datasources}

    node_list = sorted(
        [x for x in graph.nodes if graph.nodes[x]["type"] == NodeType.NODE],
        # sort so that anything with a partial match on the target is later
        key=lambda x: len(
            [x for x in identifier_map[x].partial_concepts if x in grain]
        ),
    )
    for left in node_list:
        # the constant dataset is a special case
        # and can never be on the left of a join
        if left == CONSTANT_DATASET:
            continue

        for cnode in graph.neighbors(left):
            if graph.nodes[cnode]["type"] == NodeType.CONCEPT:
                for right in graph.neighbors(cnode):
                    # skip concepts
                    if graph.nodes[right]["type"] == NodeType.CONCEPT:
                        continue
                    if left == right:
                        continue
                    identifier = [left, right]
                    joins["-".join(identifier)].add(cnode)

    final_joins_pre: List[BaseJoin] = []

    for key, join_concepts in joins.items():
        left, right = key.split("-")
        local_concepts: List[Concept] = unique(
            [c for c in concepts if c.address in join_concepts], "address"
        )
        if all([c.granularity == Granularity.SINGLE_ROW for c in local_concepts]):
            # for the constant join, make it a full outer join on 1=1
            join_type = JoinType.FULL
            local_concepts = []
        elif any(
            [
                c.address in [x.address for x in identifier_map[left].partial_concepts]
                for c in local_concepts
            ]
        ):
            join_type = JoinType.FULL
            local_concepts = [
                c for c in local_concepts if c.granularity != Granularity.SINGLE_ROW
            ]
        else:
            join_type = JoinType.LEFT_OUTER
            # remove any constants if other join keys exist
            local_concepts = [
                c for c in local_concepts if c.granularity != Granularity.SINGLE_ROW
            ]

            # if concept.keys and all([x in grain for x in concept.keys]):
            #     continue
        final_joins_pre.append(
            BaseJoin(
                left_datasource=identifier_map[left],
                right_datasource=identifier_map[right],
                join_type=join_type,
                concepts=concept_to_relevant_joins(local_concepts),
            )
        )
    final_joins = resolve_join_order(final_joins_pre)

    # this is extra validation
    non_single_row_ds = [x for x in datasources if not x.grain.abstract]
    if len(non_single_row_ds) > 1:
        for x in datasources:
            if x.grain.abstract:
                continue
            found = False
            for join in final_joins:
                if (
                    join.left_datasource.identifier == x.identifier
                    or join.right_datasource.identifier == x.identifier
                ):
                    found = True
            if not found:
                raise SyntaxError(
                    f"Could not find join for {x.identifier} with output {[c.address for c in x.output_concepts]}, all {[z.identifier for z in datasources]}"
                )
    single_row = [x for x in datasources if x.grain.abstract]
    for x in single_row:
        for join in final_joins:
            found = False
            for join in final_joins:
                if (
                    join.left_datasource.identifier == x.identifier
                    or join.right_datasource.identifier == x.identifier
                ):
                    found = True
            if not found:
                raise SyntaxError(
                    f"Could not find join for {x.identifier} with output {[c.address for c in x.output_concepts]}, all {[z.identifier for z in datasources]}"
                )
    return final_joins


def get_disconnected_components(
    concept_map: Dict[str, Set[Concept]]
) -> Tuple[int, List]:
    """Find if any of the datasources are not linked"""
    import networkx as nx

    graph = nx.Graph()
    all_concepts = set()
    for datasource, concepts in concept_map.items():
        graph.add_node(datasource, type=NodeType.NODE)
        for concept in concepts:
            # TODO: determine if this is the right way to handle things
            # if concept.derivation in (PurposeLineage.FILTER, PurposeLineage.WINDOW):
            #     if isinstance(concept.lineage, FilterItem):
            #         graph.add_node(concept.lineage.content.address, type=NodeType.CONCEPT)
            #         graph.add_edge(datasource, concept.lineage.content.address)
            #     if isinstance(concept.lineage, WindowItem):
            #         graph.add_node(concept.lineage.content.address, type=NodeType.CONCEPT)
            #         graph.add_edge(datasource, concept.lineage.content.address)
            graph.add_node(concept.address, type=NodeType.CONCEPT)
            graph.add_edge(datasource, concept.address)
            all_concepts.add(concept)
    sub_graphs = list(nx.connected_components(graph))
    sub_graphs = [
        x for x in sub_graphs if calculate_graph_relevance(graph, x, all_concepts) > 0
    ]
    return len(sub_graphs), sub_graphs



from trilogy.hooks.base_hook import BaseHook
from networkx import DiGraph


class GraphHook(BaseHook):
    def __init__(self):
        super().__init__()
        try:
            pass
        except ImportError:
            raise ImportError("GraphHook requires matplotlib and scipy to be installed")

    def query_graph_built(self, graph: DiGraph):
        from networkx import draw_kamada_kawai
        from matplotlib import pyplot as plt

        graph = graph.copy()
        nodes = [*graph.nodes]
        for node in nodes:
            if "__preql_internal" in node:
                graph.remove_node(node)
        draw_kamada_kawai(graph, with_labels=True, connectionstyle="arc3, rad = 0.1")
        # draw_spring(graph, with_labels=True, connectionstyle='arc3, rad = 0.1')
        plt.show()



from typing import Mapping, Callable, Any

from jinja2 import Template

from trilogy.core.enums import FunctionType, WindowType
from trilogy.dialect.base import BaseDialect


WINDOW_FUNCTION_MAP: Mapping[WindowType, Callable[[Any, Any, Any], str]] = {}

FUNCTION_MAP = {
    FunctionType.COUNT: lambda x: f"count({x[0]})",
    FunctionType.SUM: lambda x: f"sum({x[0]})",
    FunctionType.LENGTH: lambda x: f"length({x[0]})",
    FunctionType.AVG: lambda x: f"avg({x[0]})",
    FunctionType.LIKE: lambda x: (
        f" CASE WHEN {x[0]} like {x[1]} THEN True ELSE False END"
    ),
    FunctionType.MINUTE: lambda x: f"EXTRACT(MINUTE from {x[0]})",
    FunctionType.SECOND: lambda x: f"EXTRACT(SECOND from {x[0]})",
    FunctionType.HOUR: lambda x: f"EXTRACT(HOUR from {x[0]})",
    FunctionType.DAY_OF_WEEK: lambda x: f"EXTRACT(DAYOFWEEK from {x[0]})",
    FunctionType.DAY: lambda x: f"EXTRACT(DAY from {x[0]})",
    FunctionType.YEAR: lambda x: f"EXTRACT(YEAR from {x[0]})",
    FunctionType.MONTH: lambda x: f"EXTRACT(MONTH from {x[0]})",
    FunctionType.WEEK: lambda x: f"EXTRACT(WEEK from {x[0]})",
    FunctionType.QUARTER: lambda x: f"EXTRACT(QUARTER from {x[0]})",
    # math
    FunctionType.DIVIDE: lambda x: f"SAFE_DIVIDE({x[0]},{x[1]})",
}

FUNCTION_GRAIN_MATCH_MAP = {
    **FUNCTION_MAP,
    FunctionType.COUNT: lambda args: f"{args[0]}",
    FunctionType.SUM: lambda args: f"{args[0]}",
    FunctionType.AVG: lambda args: f"{args[0]}",
}

SQL_TEMPLATE = Template(
    """{%- if output %}
CREATE OR REPLACE TABLE {{ output.address }} AS
{% endif %}{%- if ctes %}
WITH {% for cte in ctes %}
{{cte.name}} as ({{cte.statement}}){% if not loop.last %},{% endif %}{% endfor %}{% endif %}
SELECT

{%- for select in select_columns %}
    {{ select }}{% if not loop.last %},{% endif %}{% endfor %}
{% if base %}FROM
    {{ base }}{% endif %}{% if joins %}
{% for join in joins %}
{{ join }}
{% endfor %}{% endif %}
{% if where %}WHERE
    {{ where }}
{% endif %}
{%- if group_by %}GROUP BY {% for group in group_by %}
    {{group}}{% if not loop.last %},{% endif %}{% endfor %}{% endif %}
{%- if order_by %}
ORDER BY {% for order in order_by %}
    {{ order }}{% if not loop.last %},{% endif %}
{% endfor %}{% endif %}
{%- if limit is not none %}
LIMIT {{ limit }}{% endif %}
"""
)
MAX_IDENTIFIER_LENGTH = 50


class PrestoDialect(BaseDialect):
    WINDOW_FUNCTION_MAP = {**BaseDialect.WINDOW_FUNCTION_MAP, **WINDOW_FUNCTION_MAP}
    FUNCTION_MAP = {**BaseDialect.FUNCTION_MAP, **FUNCTION_MAP}
    FUNCTION_GRAIN_MATCH_MAP = {
        **BaseDialect.FUNCTION_GRAIN_MATCH_MAP,
        **FUNCTION_GRAIN_MATCH_MAP,
    }
    QUOTE_CHARACTER = '"'
    SQL_TEMPLATE = SQL_TEMPLATE


class TrinoDialect(PrestoDialect):
    pass



from trilogy.core.enums import Purpose
from trilogy.core.models import (
    DataType,
    Parenthetical,
    ProcessedQuery,
    ShowStatement,
    SelectStatement,
    Environment,
)
from trilogy.core.functions import argument_to_purpose, function_args_to_output_purpose
from trilogy.parsing.parse_engine import (
    arg_to_datatype,
    parse_text,
)
from trilogy.constants import MagicConstants
from trilogy.dialect.base import BaseDialect


def test_in():
    _, parsed = parse_text(
        "const order_id <- 3; SELECT order_id  WHERE order_id IN (1,2,3);"
    )
    query = parsed[-1]
    right = query.where_clause.conditional.right
    assert isinstance(
        right,
        Parenthetical,
    ), type(right)
    assert right.content[0] == 1
    rendered = BaseDialect().render_expr(right)
    assert rendered.strip() == "( 1,2,3 )".strip()


def test_not_in():
    _, parsed = parse_text(
        "const order_id <- 4; SELECT order_id  WHERE order_id NOT IN (1,2,3);"
    )
    query: ProcessedQuery = parsed[-1]
    right = query.where_clause.conditional.right
    assert isinstance(
        right,
        Parenthetical,
    ), type(right)
    assert right.content[0] == 1
    rendered = BaseDialect().render_expr(right)
    assert rendered.strip() == "( 1,2,3 )".strip()


def test_is_not_null():
    _, parsed = parse_text(
        "const order_id <- 4; SELECT order_id  WHERE order_id is not null;"
    )
    query = parsed[-1]
    right = query.where_clause.conditional.right
    assert isinstance(right, MagicConstants), type(right)
    rendered = BaseDialect().render_expr(right)
    assert rendered == "null"


def test_sort():
    _, parsed = parse_text(
        "const order_id <- 4; SELECT order_id  ORDER BY order_id desc;"
    )

    _, parsed = parse_text(
        "const order_id <- 4; SELECT order_id  ORDER BY order_id DESC;"
    )
    _, parsed = parse_text(
        "const order_id <- 4; SELECT order_id  ORDER BY order_id DesC;"
    )


def test_arg_to_datatype():
    assert arg_to_datatype(1.00) == DataType.FLOAT
    assert arg_to_datatype("test") == DataType.STRING


def test_argument_to_purpose(test_environment: Environment):
    assert argument_to_purpose(1.00) == Purpose.CONSTANT
    assert argument_to_purpose("test") == Purpose.CONSTANT
    assert argument_to_purpose(test_environment.concepts["order_id"]) == Purpose.KEY
    assert (
        function_args_to_output_purpose(
            [
                "test",
                1.00,
            ]
        )
        == Purpose.CONSTANT
    )
    assert (
        function_args_to_output_purpose(
            ["test", 1.00, test_environment.concepts["order_id"]]
        )
        == Purpose.PROPERTY
    )
    unnest_env, parsed = parse_text("const random <- unnest([1,2,3,4]);")
    assert (
        function_args_to_output_purpose([unnest_env.concepts["random"]])
        == Purpose.PROPERTY
    )


def test_show(test_environment):
    _, parsed = parse_text(
        "const order_id <- 4; SHOW SELECT order_id  WHERE order_id is not null;"
    )
    query = parsed[-1]
    assert isinstance(query, ShowStatement)
    assert (
        query.content.output_components[0].address
        == test_environment.concepts["order_id"].address
    )


def test_as_transform(test_environment):
    _, parsed = parse_text("const order_id <- 4; SELECT order_id as new_order_id;")
    query = parsed[-1]
    assert isinstance(query, SelectStatement)


def test_bq_address():
    _, parsed = parse_text(
        """key user_pseudo_id int;
key event_time int;
property event_time.event_date string;

datasource fundiverse (
    event_date: event_date,
    user_pseudo_id: user_pseudo_id,
    event_time: event_time,
)
grain (event_time)
address `preqldata.analytics_411641820.events_*`
;"""
    )
    query = parsed[-1]
    assert query.address.location == "`preqldata.analytics_411641820.events_*`"


def test_purpose_and_keys():
    env, parsed = parse_text(
        """key id int;
property id.name string;

auto name_alphabetical <- row_number id order by name asc;


select
    id,
    name,
    row_number id order by name asc -> name_alphabetical_2
    ;
"""
    )

    for name in ["name_alphabetical", "name_alphabetical_2"]:
        assert name in env.concepts
        assert env.concepts[name].purpose == Purpose.PROPERTY
        assert env.concepts[name].keys == (env.concepts["id"],)


def test_output_purpose():

    env, parsed = parse_text(
        """key id int;
property id.name string;

auto name_alphabetical <- row_number id order by name asc;


rowset test<- select
    name,
    row_number id order by name asc -> name_alphabetical_2
    ;

select 
    count(test.name) -> test_name_count;
"""
    )
    # assert output_purpose == Purpose.METRIC
    for name in ["test_name_count"]:
        assert name in env.concepts
        assert env.concepts[name].purpose == Purpose.METRIC



from logging import getLogger
from dataclasses import dataclass
from enum import Enum

logger = getLogger("preql")

DEFAULT_NAMESPACE = "local"

VIRTUAL_CONCEPT_PREFIX = "_virtual"

ENV_CACHE_NAME = ".preql_cache.json"


class MagicConstants(Enum):
    NULL = "null"


NULL_VALUE = MagicConstants.NULL


# TODO: support loading from environments
@dataclass
class Config:
    strict_mode: bool = True
    human_identifiers: bool = True


CONFIG = Config()

CONFIG.strict_mode = True



from trilogy.core.models import Concept, Environment, LooseConceptList
from trilogy.utility import unique
from trilogy.core.processing.nodes import GroupNode, StrategyNode, History
from typing import List
from trilogy.core.processing.node_generators.common import (
    resolve_function_parent_concepts,
)
from trilogy.constants import logger
from trilogy.core.processing.utility import padding, create_log_lambda
from trilogy.core.processing.node_generators.common import (
    gen_enrichment_node,
)

LOGGER_PREFIX = "[GEN_GROUP_NODE]"


def gen_group_node(
    concept: Concept,
    local_optional: List[Concept],
    environment: Environment,
    g,
    depth: int,
    source_concepts,
    history: History | None = None,
):
    # aggregates MUST always group to the proper grain
    # except when the
    parent_concepts: List[Concept] = unique(
        resolve_function_parent_concepts(concept), "address"
    )
    logger.info(
        f"{padding(depth)}{LOGGER_PREFIX} parent_concepts are {[x.address for x in parent_concepts]} from group grain {concept.grain}"
    )

    # if the aggregation has a grain, we need to ensure these are the ONLY optional in the output of the select
    output_concepts = [concept]

    if concept.grain and len(concept.grain.components_copy) > 0:
        grain_components = (
            concept.grain.components_copy if not concept.grain.abstract else []
        )
        parent_concepts += grain_components
        output_concepts += grain_components

    if parent_concepts:
        logger.info(
            f"{padding(depth)}{LOGGER_PREFIX} fetching group node parents {LooseConceptList(concepts=parent_concepts)}"
        )
        parent_concepts = unique(parent_concepts, "address")
        parent = source_concepts(
            mandatory_list=parent_concepts,
            environment=environment,
            g=g,
            depth=depth,
            history=history,
        )
        if not parent:
            logger.info(
                f"{padding(depth)}{LOGGER_PREFIX} group by node parents unresolvable"
            )
            return None
        parents: List[StrategyNode] = [parent]
    else:
        parents = []

    # the keys we group by
    # are what we can use for enrichment
    group_key_parents = concept.grain.components_copy

    group_node = GroupNode(
        output_concepts=output_concepts,
        input_concepts=parent_concepts,
        environment=environment,
        g=g,
        parents=parents,
        depth=depth,
    )

    # early exit if no optional
    if not local_optional:
        return group_node
    logger.info(f"{padding(depth)}{LOGGER_PREFIX} group node requires enrichment")
    return gen_enrichment_node(
        group_node,
        join_keys=group_key_parents,
        local_optional=local_optional,
        environment=environment,
        g=g,
        depth=depth,
        source_concepts=source_concepts,
        log_lambda=create_log_lambda(LOGGER_PREFIX, depth, logger),
    )



from trilogy.core.enums import BooleanOperator, Purpose, JoinType, ComparisonOperator
from trilogy.core.models import (
    CTE,
    Grain,
    QueryDatasource,
    Conditional,
    SelectStatement,
    Environment,
    Address,
    UndefinedConcept,
    BaseJoin,
    Comparison,
)


def test_cte_merge(test_environment, test_environment_graph):
    datasource = list(test_environment.datasources.values())[0]
    outputs = [c.concept for c in datasource.columns]
    output_map = {
        c.address: {
            datasource,
        }
        for c in outputs
    }
    a = CTE(
        name="test",
        output_columns=[outputs[0]],
        grain=Grain(),
        source=QueryDatasource(
            input_concepts=[outputs[0]],
            output_concepts=[outputs[0]],
            datasources=[datasource],
            grain=Grain(),
            joins=[],
            source_map={outputs[0].address: {datasource}},
        ),
        source_map={c.address: datasource.identifier for c in outputs},
    )
    b = CTE(
        name="testb",
        output_columns=outputs,
        grain=Grain(),
        source=QueryDatasource(
            input_concepts=outputs,
            output_concepts=outputs,
            datasources=[datasource],
            grain=Grain(),
            joins=[],
            source_map=output_map,
        ),
        source_map={c.address: datasource.identifier for c in outputs},
    )

    merged = a + b
    assert merged.output_columns == outputs


def test_concept(test_environment, test_environment_graph):
    test_concept = list(test_environment.concepts.values())[0]
    new = test_concept.with_namespace("test")
    assert (
        new.namespace == ("test" + "." + test_concept.namespace)
        if test_concept.namespace != "local"
        else "test"
    )


def test_conditional(test_environment, test_environment_graph):
    test_concept = list(test_environment.concepts.values())[-1]

    condition_a = Conditional(
        left=test_concept, right=test_concept, operator=BooleanOperator.AND
    )
    condition_b = Conditional(
        left=test_concept, right=test_concept, operator=BooleanOperator.AND
    )
    merged = condition_a + condition_b
    assert merged == condition_a

    test_concept_two = list(test_environment.concepts.values())[-2]
    condition_c = Conditional(
        left=test_concept, right=test_concept_two, operator=BooleanOperator.AND
    )
    merged_two = condition_a + condition_c
    assert merged_two.left == condition_a
    assert merged_two.right == condition_c
    assert merged_two.operator == BooleanOperator.AND


def test_grain(test_environment):
    grains = [Grain()] * 3
    sum(grains)
    oid = test_environment.concepts["order_id"]
    pid = test_environment.concepts["product_id"]
    cid = test_environment.concepts["category_id"]
    cname = test_environment.concepts["category_name"]

    x = Grain(components=[oid, pid])
    y = Grain(components=[pid, cid])
    z = Grain(components=[cid])
    z2 = Grain(components=[cid, cname])

    assert x.intersection(y) == Grain(components=[pid])
    assert x.union(y) == Grain(components=[oid, pid, cid])

    assert z.isdisjoint(x)
    assert z.issubset(y)

    assert z2 == z, f"Property should be removed from grain ({z.set}) vs {z2.set}"


def test_select(test_environment: Environment):
    oid = test_environment.concepts["order_id"]
    pid = test_environment.concepts["product_id"]
    cid = test_environment.concepts["category_id"]
    cname = test_environment.concepts["category_name"]
    x = SelectStatement(selection=[oid, pid, cid, cname])
    ds = x.to_datasource(
        test_environment.namespace, "test", address=Address(location="test")
    )

    assert ds.grain == Grain(components=[oid, pid, cid])


def test_undefined(test_environment: Environment):
    x = UndefinedConcept(
        name="test",
        datatype="int",
        purpose=Purpose.CONSTANT,
        grain=Grain(),
        namespace="test",
        environment=test_environment.concepts,
    )

    y = x.with_select_grain(Grain(components=[test_environment.concepts["order_id"]]))

    assert y.grain == Grain(components=[test_environment.concepts["order_id"]])

    z = x.with_default_grain()

    assert z.grain == Grain()


def test_base_join(test_environment: Environment):
    exc: SyntaxError | None = None
    try:
        BaseJoin(
            left_datasource=test_environment.datasources["revenue"],
            right_datasource=test_environment.datasources["revenue"],
            concepts=[test_environment.concepts["product_id"]],
            join_type=JoinType.RIGHT_OUTER,
        )
    except Exception as exc2:
        exc = exc2
        pass
    assert isinstance(exc, SyntaxError)

    x = BaseJoin(
        left_datasource=test_environment.datasources["revenue"],
        right_datasource=test_environment.datasources["products"],
        concepts=[
            test_environment.concepts["product_id"],
            test_environment.concepts["category_name"],
        ],
        join_type=JoinType.RIGHT_OUTER,
        filter_to_mutual=True,
    )

    assert x.concepts == [test_environment.concepts["product_id"]]

    exc3: SyntaxError | None = None
    try:
        x = BaseJoin(
            left_datasource=test_environment.datasources["revenue"],
            right_datasource=test_environment.datasources["category"],
            concepts=[
                test_environment.concepts["product_id"],
                test_environment.concepts["category_name"],
            ],
            join_type=JoinType.RIGHT_OUTER,
            filter_to_mutual=True,
        )
    except Exception as exc4:
        exc3 = exc4
        pass
    assert isinstance(exc3, SyntaxError)


def test_comparison():
    try:
        Comparison(left=1, right="abc", operator=ComparisonOperator.EQ)
    except Exception as exc:
        assert isinstance(exc, SyntaxError)
